# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --

"""
   Variables used in this module:
    :ncore:     number of frozen core orbitals
    :nocc:      number of occupied orbitals in the principle configuration
    :nacto:     number of active occupied orbitals in the principle configuration
    :nvirt:     number of virtual orbitals in the principle configuration
    :nactv:     number of active virtual orbitals in the principle configuration
    :nbasis:    total number of basis functions
    :nact:      total number of active orbitals (nacto+nactv)
    :e_ci:      eigenvalues of CI Hamiltonian (IOData container attribute)
    :civ:       eigenvectors of CI Hamiltonian (IOData container attribute)

    Indexing convention:
     :i,j,k,..: occupied orbitals of principle configuration
     :a,b,c,..: virtual orbitals of principle configuration
"""

from abc import ABC, abstractmethod
import numpy as np

from pybest.cache import Cache
from pybest import filemanager
from pybest.corrections.rci_corrections import RCICorrections
from pybest.utils import check_options
from pybest.io.iodata import IOData
from pybest.featuredlists import OneBodyHamiltonian, TwoBodyHamiltonian
from pybest.solvers.davidson import Davidson
from pybest.log import log
from pybest.utils import (
    unmask_orb,
    unmask,
)
from pybest.io.checkpoint import CheckPoint
from pybest.linalg.dense import DenseLinalgFactory
from pybest.orbital_utils import transform_integrals, split_core_active
from pybest.linalg.base import TwoIndex, FourIndex
from pybest.exceptions import ArgumentError


class RCI(ABC):
    # pylint: disable=too-many-instance-attributes
    """Configuration Interaction base class"""

    long_name = ""
    acronym = ""
    reference = ""

    def __init__(self, lf, occ_model):
        """
        **Arguments:**

        lf
             A LinalgFactory instance.

        occ_model
             Occupation model

        **Optional arguments:**

        ncore
             Number of frozen core orbitals. If provided, the core orbitals
             are excluded from the CI optimization.

        """
        #
        # Variables
        #
        self._lf = lf
        self._denself = DenseLinalgFactory(lf.default_nbasis)
        self._nocc = occ_model.nocc[0]

        self._occ_model = occ_model
        self._nacto = occ_model.nocc[0] - occ_model.ncore[0]
        self._ncore = occ_model.ncore[0]
        if self._nacto <= 0 or self._nacto > occ_model.nocc[0]:
            raise ValueError(
                f"Impossible number of frozen core orbitals! You have chosen {ncore} "
                f"frozen core orbitals out of {occ_model.nocc[0]} occupied orbitals"
            )

        self._nbasis = lf.default_nbasis
        self._nact = lf.default_nbasis - self._ncore
        self._nvirt = self._nbasis - self._nocc
        self._nactv = self._nact - self._nacto

        if self._nvirt <= 0 or self._nvirt > (
            lf.default_nbasis - occ_model.nocc[0]
        ):
            raise ValueError(
                f"Impossible number of virtual orbitals: {self._nvirt}!"
            )

        if self._nactv <= 0 or self._nactv > (
            lf.default_nbasis - occ_model.nocc[0]
        ):
            raise ValueError(
                f"Impossible number of active virtual orbitals: {self._nactv}!"
            )

        self._e_core = 0.0
        self._cache = Cache()
        self._checkpoint = CheckPoint({})
        self._checkpoint_fn = f"checkpoint_{self.acronym}.h5"
        self._todisk = False
        self._tco = "td"
        self._nroot = 1
        self._dimension = 0
        self._print_csf = False
        # Include occupation model into checkpoint file (IOData container)
        self.checkpoint.update("occ_model", self.occ_model)

    #
    #  Properties
    #
    @property
    def e_core(self):
        """Core energy"""
        return self._e_core

    @property
    def occ_model(self):
        """The occupation model."""
        return self._occ_model

    @property
    def nactv(self):
        """The number of active virtual orbitals"""
        return self._nactv

    @property
    def nacto(self):
        """The number of active occupied orbitals"""
        return self._nacto

    @property
    def ncore(self):
        """The number of core orbitals"""
        return self._ncore

    @property
    def nact(self):
        """The number of active basis functions"""
        return self._nact

    @property
    def rci_corrections(self):
        """Instance of RCICorrections class"""
        return self._rci_corrections

    @rci_corrections.setter
    def rci_corrections(self, new):
        self._rci_corrections = new

    @property
    def threshold_c_0(self):
        """The threshold for Davidson-type corrections.
        Forwarded to RCICorrections module."""
        return self._threshold_c_0

    @threshold_c_0.setter
    def threshold_c_0(self, new):
        self._threshold_c_0 = new

    @property
    def threshold(self):
        """The threshold for printing CI coefficients"""
        return self._threshold

    @threshold.setter
    def threshold(self, new):
        self._threshold = new

    @property
    def lf(self):
        """The linalg factory"""
        return self._lf

    @property
    def nocc(self):
        """The number of occupied orbitals"""
        return self._nocc

    @property
    def nvirt(self):
        """The number of virtual orbitals"""
        return self._nvirt

    @property
    def nbasis(self):
        """The total number of basis functions"""
        return self._nbasis

    @property
    def denself(self):
        """The dense linalg factory"""
        return self._denself

    @property
    def tco(self):
        """The tensor contraction engine used"""
        return self._tco

    @tco.setter
    def tco(self, new):
        self._tco = new

    @property
    def size_consistency_correction(self):
        """A boolean variable (default: True):
        -True: perform Davidson-type corrections"""
        return self._size_consistency_correction

    @size_consistency_correction.setter
    def size_consistency_correction(self, new):
        if not isinstance(new, bool):
            raise ArgumentError(
                "Unkown type for keyword correction. Boolean type required."
            )
        self._size_consistency_correction = new

    @property
    def davidson(self):
        """A boolean variable (default: True):
        -True: perform Davidson diagonalization
        -False: perform exact diagonalization."""
        return self._davidson

    @davidson.setter
    def davidson(self, new):
        if not isinstance(new, bool):
            raise ArgumentError(
                "Unkown type for keyword davidson. Boolean type required."
            )
        self._davidson = new

    @property
    def print_csf(self):
        """A boolean variable (default: False):
        -True: print the results in CSF basis
        -False: print the results in SD basis."""
        return self._print_csf

    @print_csf.setter
    def print_csf(self, new):
        self._print_csf = new

    @property
    def e_ref(self):
        """The total energy of the reference wave function."""
        return self._e_ref

    @e_ref.setter
    def e_ref(self, new):
        self._e_ref = new

    @property
    def cache(self):
        """A Cache instance."""
        return self._cache

    @property
    def checkpoint(self):
        """The iodata container that contains all data dump to disk."""
        return self._checkpoint

    @checkpoint.setter
    def checkpoint(self, new):
        self._checkpoint = new

    @property
    def checkpoint_fn(self):
        """The filename that will be dumped to disk."""
        return self._checkpoint_fn

    @property
    def todisk(self):
        """Dumping the eigenvectors to disk
        (default: False, True-currently not supported)"""
        return self._todisk

    @todisk.setter
    def todisk(self, new):
        self._todisk = new

    @property
    def csf(self):
        """Configuration State Functions. If True, the CSF's are used during
        optimization. Default: False.
        """
        return self._csf

    @csf.setter
    def csf(self, new):
        if not isinstance(new, bool):
            raise ArgumentError(
                "Unkown type for keyword csf. Boolean type required."
            )
        self._csf = new

    @property
    def nroot(self):
        """The number of targeted states."""
        return self._nroot

    @nroot.setter
    @abstractmethod
    def nroot(self, new):
        raise NotImplementedError

    @property
    @abstractmethod
    def ci_method(self):
        """Instance of SD- or CSF-type class"""

    @property
    def dimension(self):
        """The dimension of the Hamiltonian matrix."""
        return self._dimension

    @dimension.setter
    @abstractmethod
    def dimension(self, new=None):
        raise NotImplementedError

    @abstractmethod
    def collect_data(self, index, data, evecsj):
        """Collects the data and prepares them for printing"""

    @abstractmethod
    def set_hamiltonian(self, mo1, mo2):
        """Computation of auxiliary matrices."""

    # pylint: disable=too-many-arguments
    @abstractmethod
    def init_cache(
        self, select, dim0=None, dim1=None, dim2=None, dim3=None, nvec=None
    ):
        """Initialize auxiliary matrices"""

    @abstractmethod
    def calculate_exact_hamiltonian(self):
        """Calculates the exact Hamiltonian for the chosen model."""

    @abstractmethod
    def printer(self):
        """Printing the results."""

    #
    # FUNCTIONS
    #
    def print_info(self):
        """Prints information about the CI calculations."""
        if log.do_medium:
            log(" ")
            log(f"Entering {self.long_name} ")
            log(" ")
            log.hline("~")
            log(f"{self.acronym} framework selected")
            log.hline("~")
            log("OPTIMIZATION PARAMETERS:")
            log(f"CI model:                          {self.acronym}")
            log(f"Reference function:                {self.reference}")
            log(f"Number of frozen cores:            {self.ncore}")
            log(f"Number of active occupied:         {self.nacto}")
            log(f"Number of active virtuals:         {self.nactv}")
            log(f"Number of targeted roots:          {self.nroot}")
            log(f"Total number of roots:             {self.dimension}")
            log(f"Spin-adapted configurations (CSF): {self.csf}")
            log(
                f"Davidson-type corrections:         {self.size_consistency_correction}"
            )
            if self.davidson:
                log("Diagonalization:                   Davidson")
            else:
                log("Diagonalization:                   Exact Diagonalization")
            log.hline("~")

    def get_range(self, string, start=0):
        """Returns dictionary with keys beginX, endX, begin(X+1), etc.
        *  string - any sequence of 'o' (occupied) and 'v' (virtual)
        """
        range_ = {}
        ind = start
        for char in string:
            if char == "o":
                range_["begin" + str(ind)] = 0
                range_["end" + str(ind)] = self.nacto
            elif char == "v":
                range_["begin" + str(ind)] = self.nacto
                range_["end" + str(ind)] = self.nact
            elif char == "V":
                range_["begin" + str(ind)] = 0
                range_["end" + str(ind)] = self.nactv
            elif char == "n":
                range_["begin" + str(ind)] = 0
                range_["end" + str(ind)] = self.nact
            else:
                raise ArgumentError(
                    f"Do not know how to handle choice {char}."
                )
            ind += 1
        return range_

    def read_input(self, **kwargs):
        """Reads and sets up input parameters."""
        for name in kwargs:
            check_options(
                name,
                name,
                "threshold",
                "maxiter",
                "nroot",
                "e_ref",
                "tolerance",
                "tolerancev",
                "nguessv",
                "tco",
                "maxvectors",
                "davidson",
                "scc",
                "threshold_c_0",
                "print_csf",
            )
        #
        # Set default keyword arguments:
        #
        self.threshold = kwargs.get("threshold", 0.1)
        self.davidson = kwargs.get("davidson", True)
        self.print_csf = kwargs.get("print_csf", False)
        self.size_consistency_correction = kwargs.get("scc", True)
        self.threshold_c_0 = kwargs.get("threshold_c_0", 0.3)
        self.nroot = kwargs.get("nroot", 1)
        if self.nroot <= 0:
            raise ValueError("Number of roots must be larger than 0!")

    def unmask_args(self, *args, **kwargs):
        """Resolve arguments passed to function call"""
        #
        # olp
        #
        olp = unmask("olp", *args, **kwargs)
        if olp is None:
            raise ArgumentError("Cannot find overlap integrals.")
        self.checkpoint.update("olp", olp)
        #
        # orb
        #
        orbs = unmask_orb(*args, **kwargs)
        if orbs:
            orbs = orbs[0]
            self.checkpoint.update("orb_a", orbs.copy())
        else:
            raise ArgumentError("Cannot find orbitals.")
        #
        # 1-e ints and 2-e ints
        #
        one = self.lf.create_two_index(label="one")
        for arg in args:
            if isinstance(arg, TwoIndex):
                if arg.label in OneBodyHamiltonian:
                    one.iadd(arg)
            elif isinstance(arg, FourIndex):
                if arg.label in TwoBodyHamiltonian:
                    two = arg
        return one, two, orbs

    def build_guess_vectors(self, nguess, todisk, *args):
        """Used by davidson module to construct guess"""

        bvector = []
        hdiag = args[0]
        sortedind = hdiag.sort_indices(True)
        dim = self.dimension
        #
        # Construct Guess vectors according to hdiag
        #
        b = self.lf.create_one_index(dim)
        count = 0
        for ind in range(nguess):
            if ind >= dim:
                if log.do_medium:
                    log.warn(f"Maximum number of guess vectors reached: {dim}")
                break
            b.clear()
            b.set_element(sortedind[ind], 1)
            if todisk:
                b_v = IOData(vector=b)
                filename = filemanager.temp_path(f"bvector_{ind}.h5")
                b_v.to_file(filename)
                count += 1
            else:
                # have to append a copy
                bvector.append(b.copy())
        if todisk:
            return None, count
        return bvector, len(bvector)

    def __call__(self, *args, **kwargs):
        """Solve the symmetric eigenvalue problem in CI method

        Currently supported CI techniques of solving:
         * Slater Determinant (SD)
         * Configuration State Function (CSF)

        **Arguments:**

        args
            Contains various arguments: One- and two-body integrals
            (some Hamiltonian matrix elements) expressed in the AO basis,
            an IOData container containing the wave function information
            of the reference state (the AO/MO coefficients, CC amplitudes,
            etc.)

        **Keywords:**

            Contains the following keyword arguments:
             * tolerance:   tolerance for energies (default 1e-6)
             * tolerancev:  tolerance for eigenvectors (default 1e-5)
             * threshold:   printing threshold for contributions of CI wave function (default 0.1)
             * maxiter:     maximum number of iterations (default 200)
             * nroot :      the number of targeted states (default depend on model)
             * nguessv:     total number of guess vectors (default (nroot-1)*4+1)
             * maxvectors:  maximum number of Davidson vectors (default (nroot-1)*10)
             * davidson:    A boolean variable (default True):
                                -True: perform Davidson diagonalization
                                -False: perform exact diagonalization.
             * csf:         (boolean) take spin-adapted configurations (configuration
                            state functions) as basis instead of working with
                            Slater determinants (default False)
            * scc:          (boolean) True option will calculate the size-consistency
                            correction (default True)
            * threshold_c_0:
                            threshold that helps verifing the accuracy of
                            Davidson-type corrections
        **Returns**

            An IOData container containing all eigenvalues and eigenvectors
            for targeted states.
        """
        #
        # Read input:
        #
        self.read_input(**kwargs)
        #
        # Setup:
        #
        one, two, orbs = self.unmask_args(*args, **kwargs)
        self.e_ref = unmask("e_tot", *args, **kwargs) or unmask(
            "e_ref", *args, **kwargs
        )
        #
        # Print Info:
        #
        self.print_info()
        #
        # Transform of integrals:
        #
        if self.ncore > 0:
            cas = split_core_active(
                one,
                two,
                orbs,
                e_core=self.e_core,
                ncore=self.ncore,
                nactive=self.nact,
            )
            mo1 = cas.one
            mo2 = cas.two
        else:
            t_i = transform_integrals(one, two, orbs)
            (mo1,) = t_i.one
            (mo2,) = t_i.two
        #
        # Auxiliary objects
        #
        self.clear_cache()
        self.set_hamiltonian(mo1, mo2)
        del mo1, mo2

        self.diagonalize(**kwargs)
        self.clear_cache()

        if log.do_medium:
            self.printer()

        return self.checkpoint()

    def diagonalize(self, **kwargs):
        # pylint: disable=eval-used
        """
        Solve for CI Hamiltonian, returns eigenvalues and eigenvectors (stored in
        IOData container of class)

        **Arguments:**

        args
             Contains various arguments: One- and two-body integrals
             (some Hamiltonian matrix elements) expressed in the AO basis,
             an IOData container containing the wave function information
             of the reference state (the AO/MO coefficients, CC amplitudes,
             etc.)

        **Keywords:**

             * tolerance:   tolerance for energies (default 1e-6)
             * tolerancev:  tolerance for eigenvectors (default 1e-5)
             * threshold:   printing threshold for contributions of CI wave function  (default 0.1)
             * maxiter:     maximum number of iterations (default 200)
             * nroot :      the number of targeted states (default depend on model)
             * nguessv:     total number of guess vectors (default (nroot-1)*4+1)
             * maxvectors:  maximum number of Davidson vectors (default (nroot-1)*10)
             * davidson:    A boolean variable (default True):
                                -True: perform Davidson diagonalization
                                -False: perform exact diagonalization.
        """
        #
        # Set default keyword arguments:
        #
        tol = kwargs.get("tolerance", 1e-6)
        tolv = kwargs.get("tolerancev", 1e-4)
        maxiter = kwargs.get("maxiter", 200)
        nguessv = kwargs.get("nguessv", self.nroot * 4)
        maxvectors = kwargs.get("maxvectors", self.nroot * 10)
        #
        # Diagonalization
        #
        if self.davidson:
            davidson = Davidson(
                self.lf,
                self.nroot,
                nguess=nguessv,
                maxiter=maxiter,
                maxvectors=maxvectors,
                tolerance=tol,
                tolerancev=tolv,
                todisk=self.todisk,
                skipgs=False,
            )
            e_ci, eigen_vector_real = davidson(self.ci_method)
        else:
            hamiltonian = self.calculate_exact_hamiltonian()
            e_ci, eigen_vector_real = np.linalg.eigh(hamiltonian.array)
        #
        # Do Checkpoint
        #
        self.checkpoint.to_file(self.checkpoint_fn)
        self.checkpoint.update("e_ci", e_ci[: self.nroot])
        self.checkpoint.update("civ", eigen_vector_real[:, : self.nroot])
        self.checkpoint.update("e_ref", self.e_ref)
        #
        # Perform a size-consistency correction
        #
        if self.size_consistency_correction:
            self.rci_corrections = RCICorrections(self.nacto)
            self.rci_corrections(
                eigen_vector_real[0][0],
                e_ci[0],
                threshold_c_0=self.threshold_c_0,
                display=False,
                e_ref=self.e_ref,
                acronym=self.acronym,
            )

    def get_index_s(self, index):
        """Get index for single excitation."""
        b = index % self.nactv
        j = ((index - b) / self.nactv) % self.nacto
        return int(j), int(b)

    def from_cache(self, select):
        """Load the auxiliary matrices."""
        return self._cache.load(select)

    def clear_cache(self):
        """Clear the auxiliary matrices."""
        self._cache.clear(tags="m", dealloc=True)
