# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --


import pytest
import numpy as np

from pybest.exceptions import ArgumentError, ElectronCountError
from pybest.gbasis import get_gobasis
from pybest.context import context
from pybest.linalg.dense import DenseLinalgFactory
from pybest.occ_model import FixedOccModel


###############################################################################
## Tests for AufbauOccModel ###################################################
###############################################################################


test_fixed = [
    # basis, molecule, kwargs (charge, #unpaired electrons (alpha), ncore)
    (
        "cc-pvdz",
        "test/water.xyz",
        {"occ_a": np.array([1, 1, 1, 1, 1]), "nel": 10},
        {
            "occ": [[1, 1, 1, 1, 1]],
            "charge": 0,
            "nel": 10,
            "nbasis": [24],
            "nact": [24],
            "nocc": [5],
            "nvirt": [19],
            "nacto": [5],
            "nactv": [19],
            "ncore": [0],
            "occ_array": [np.array([1, 1, 1, 1, 1])],
            "nspin": [5],
        },
    ),
    (
        "cc-pvdz",
        "test/water.xyz",
        {"occ_a": np.array([1, 1, 1, 1, 0, 1]), "nel": 10},
        {
            "occ": [[1, 1, 1, 1, 0, 1]],
            "charge": 0,
            "nel": 10,
            "nbasis": [24],
            "nact": [24],
            "nocc": [6],
            "nvirt": [18],
            "nacto": [6],
            "nactv": [18],
            "ncore": [0],
            "occ_array": [np.array([1, 1, 1, 1, 0, 1])],
            "nspin": [5],
        },
    ),
    (
        "cc-pvdz",
        "test/water.xyz",
        {"occ_a": np.array([1, 1, 1, 1, 0, 1]), "nel": 10, "ncore": 1},
        {
            "occ": [[1, 1, 1, 1, 0, 1]],
            "charge": 0,
            "nel": 10,
            "nbasis": [24],
            "nact": [23],
            "nocc": [6],
            "nvirt": [18],
            "nacto": [5],
            "nactv": [18],
            "ncore": [1],
            "occ_array": [np.array([1, 1, 1, 1, 0, 1])],
            "nspin": [5],
        },
    ),
    (
        "cc-pvdz",
        "test/water.xyz",
        {
            "occ_a": np.array([1, 1, 1, 1, 0, 1]),
            "occ_b": np.array([1, 1, 1, 1, 0, 0, 1]),
            "nel": 10,
            "ncore": 0,
        },
        {
            "occ": [[1, 1, 1, 1, 0, 1], [1, 1, 1, 1, 0, 0, 1]],
            "charge": 0,
            "nel": 10,
            "nbasis": [24, 24],
            "nact": [24, 24],
            "nocc": [6, 7],
            "nvirt": [18, 17],
            "nacto": [6, 7],
            "nactv": [18, 17],
            "ncore": [0, 0],
            "occ_array": [
                np.array([1, 1, 1, 1, 0, 1]),
                np.array([1, 1, 1, 1, 0, 0, 1]),
            ],
            "nspin": [5, 5],
        },
    ),
]


test_instance = ["gobasis", "lf"]

###############################################################################
## Tests for FixedOccModel ####################################################
###############################################################################


@pytest.mark.parametrize("basis,mol,kwargs,expected", test_fixed)
@pytest.mark.parametrize("instance", test_instance)
def test_occ_fixed(basis, mol, kwargs, expected, instance):
    """Test Fixed Aufbau model for gobasis, nocc_a/nocc_b, and ncore arguments as
    input."""
    fn_xyz = context.get_fn(mol)
    gobasis = get_gobasis(basis, fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(gobasis.nbasis)
    # either restricted (alpha=0) or unrestricted orbitals (alpha>0)
    occ_b = kwargs.get("occ_b", None)
    norbs = 1 if occ_b is None else 2
    orb = [lf.create_orbital() for n in range(norbs)]

    occ_model = FixedOccModel(eval(instance), **kwargs)
    occ_model.assign_occ_reference(*orb)

    for key, value in expected.items():
        if key in ["occ", "occ_array"]:
            # check occupation numbers of each orbital (alpha, beta)
            assert len(orb) == len(value)
            for orb_, el in zip(orb, value):
                assert abs(orb_.occupations[: len(el)] - el).max() < 1e-10
        else:
            # check all other attributes
            assert getattr(occ_model, key) == value


#
# Invalid calls, all test should raise an error
#

test_wrong_arguments = [
    (
        # nel, needed for lf tests
        {},
        ArgumentError,
    ),  # missing arguments occ_a
    (
        {"occ_b": np.array([1, 1, 1, 1, 1])},
        ArgumentError,
    ),  # missing arguments occ_a
    (
        {"occ_a": np.array([1, 1, 1, 1, 0]), "nel": 10},
        ElectronCountError,
    ),  # occ and nel do not agree
    (
        {
            "occ_a": np.array([1, 1, 1, 1, 0, 1]),
            "occ_b": np.array([1, 1, 0, 0, 0, 1]),
            "nel": 9,
            "charge": 1,
        },
        ElectronCountError,
    ),  # nocc and nel do not agree
    (
        {
            "occ_a": np.array([1, 1, 1, 1, 0, 1]),
            "nel": 9,
            "charge": 1,
        },
        ElectronCountError,
    ),  # nocc and nel do not agree
    (
        {
            "occ_a": np.array([1, 1, 1, 1, 0, 1]),
            "nel": 8,
        },
        ElectronCountError,
    ),  # nocc and nel do not agree
]


@pytest.mark.parametrize("kwargs,error", test_wrong_arguments)
@pytest.mark.parametrize("instance", test_instance)
def test_occ_fixed_arguments_gobasis(kwargs, error, instance):
    """Test Fixed occupation model for gobasis, nocc_a/nocc_b, and ncore arguments as
    input.
    """
    fn_xyz = context.get_fn("test/water.xyz")
    gobasis = get_gobasis("cc-pvdz", fn_xyz, print_basis=False)
    lf = DenseLinalgFactory(gobasis.nbasis)

    with pytest.raises(error):
        assert FixedOccModel(eval(instance), **kwargs)
