# -*- coding: utf-8 -*-
# PyBEST: Pythonic Black-box Electronic Structure Tool
# Copyright (C) 2016-- The PyBEST Development Team
#
# This file is part of PyBEST.
#
# PyBEST is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# PyBEST is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>
#
# --
# pylint: disable=too-many-arguments

import numpy as np
import pytest

from pybest.geminals import ROOpCCD
from pybest.exceptions import ArgumentError
from pybest.linalg import DenseLinalgFactory, CholeskyLinalgFactory
from pybest.wrappers.hf import RHF
from pybest.wrappers.multipole import compute_dipole_moment
from pybest.wrappers.tests.common import Molecule
import pybest.gbasis.gobasis as pybest_basis


if pybest_basis.PYBEST_CHOLESKY_ENABLED:
    linalg_set = [DenseLinalgFactory, CholeskyLinalgFactory]
else:
    linalg_set = [DenseLinalgFactory]


test_data_hf = [
    (
        RHF,
        "h2o",
        0,
        -75.943961863294,
        {
            "dipole_hf": [0.04129661, 2.25353156, 0.00000000],
        },
    ),
]

test_data_oopccd = [
    (
        ROOpCCD,
        "h2o",
        {
            "dipole_pccd": [0.2797343, 1.9267571, 0.00000000],
            "dipole_lccsd": [-0.01807735, 2.20762975, 0.00000000],
        },
    ),
]
# Test can be extended to more molecules or to UHF in the future
@pytest.mark.parametrize("cls,mol,alpha,e_tot_hf,expected", test_data_hf)
@pytest.mark.parametrize("linalg", linalg_set)
def test_dipole_hf(cls, mol, alpha, e_tot_hf, expected, linalg):
    """Test HF dipole moment."""

    mol_ = Molecule(linalg, "cc-pvdz", f"test/{mol}.xyz", alpha=alpha)
    # Do RHF/UHF using specific DIIS and tight threshold
    mol_.do_scf(cls)

    assert abs(e_tot_hf - mol_.scf.e_tot) <= 1e-8
    assert abs(e_tot_hf - mol_.scf.e_ref) <= 1e-8

    mol_.compute_dipole()
    dipole_hf = np.array(compute_dipole_moment(mol_.dipole, mol_.scf))

    assert (abs(dipole_hf - expected["dipole_hf"]) < 1e-3).all()


# Test can be extended to more molecules or to UHF in the future
@pytest.mark.parametrize("cls,mol,alpha,e_tot_hf,expected", test_data_hf)
@pytest.mark.parametrize("linalg", linalg_set)
def test_dipole_hf_args(cls, mol, alpha, e_tot_hf, expected, linalg):
    """Test various args in SCF wrapper for dipole moment."""

    mol_ = Molecule(linalg, "cc-pvdz", f"test/{mol}.xyz", alpha=alpha)
    # Do RHF/UHF using specific DIIS and tight threshold
    mol_.do_scf(cls)

    assert abs(e_tot_hf - mol_.scf.e_tot) <= 1e-8
    assert abs(e_tot_hf - mol_.scf.e_ref) <= 1e-8

    mol_.compute_dipole()
    dipole_hf = np.array(compute_dipole_moment(mol_.dipole, mol_.scf))

    dm = mol_.scf.dm_1.copy()
    del mol_.scf.dm_1

    # missing dm
    with pytest.raises(ArgumentError):
        compute_dipole_moment(mol_.dipole, mol_.scf)

    mol_.scf.dm_1 = dm
    del mol_.scf.orb_a

    # missing orb
    with pytest.raises(ArgumentError):
        compute_dipole_moment(mol_.dipole, mol_.scf, mo=True)

    # different scaling factor
    dipole_hf = np.array(
        compute_dipole_moment(mol_.dipole, mol_.scf, scale=1.0)
    )

    assert (abs(dipole_hf * 2.0 - expected["dipole_hf"]) < 1e-3).all()


# Test OOpCCD-LCCSD dipole moment in Debyes
@pytest.mark.slow
@pytest.mark.parametrize("cls,mol,expected", test_data_oopccd)
@pytest.mark.parametrize("linalg", linalg_set)
def test_dipole_oopccd(cls, mol, expected, linalg):
    """Test dipole moment from OOpCCD and OOpCCD-LCCSD ."""

    mol_ = Molecule(linalg, "cc-pvdz", f"test/{mol}.xyz", alpha=0)
    # Do RHF using specific DIIS and tight threshold
    mol_.do_scf(RHF)
    mol_.do_pccd(cls)
    mol_.compute_dipole()
    dipole_pccd = np.array(
        compute_dipole_moment(mol_.dipole, mol_.pccd, mo=True)
    )

    assert (abs(dipole_pccd - expected["dipole_pccd"]) < 1e-3).all()

    mol_.do_pccd_lccsd()
    mol_.compute_dipole()
    dipole_pccd_lccsd = np.array(
        compute_dipole_moment(mol_.dipole, mol_.pccd_lccsd, mo=True)
    )

    assert (abs(dipole_pccd_lccsd - expected["dipole_lccsd"]) < 1e-3).all()
